@@ -56,7 +56,7 @@ static int diff_insert_delta(
 
 	if (diff->opts.notify_cb) {
 		error = diff->opts.notify_cb(
-			diff, delta, matched_pathspec, diff->opts.notify_payload);
+			diff, delta, matched_pathspec, diff->opts.payload);
 
 		if (error) {
 			git__free(delta);
@@ -79,7 +79,7 @@ static bool diff_pathspec_match(
 	git_diff *diff,
 	const git_index_entry *entry)
 {
-	bool disable_pathspec_match = 
+	bool disable_pathspec_match =
 		DIFF_FLAG_IS_SET(diff, GIT_DIFF_DISABLE_PATHSPEC_MATCH);
 
 	/* If we're disabling fnmatch, then the iterator has already applied
@@ -131,7 +131,7 @@ static int diff_delta__from_one(
 	if (status == GIT_DELTA_UNTRACKED &&
 		DIFF_FLAG_ISNT_SET(diff, GIT_DIFF_INCLUDE_UNTRACKED))
 		return 0;
-	
+
 	if (status == GIT_DELTA_UNREADABLE &&
 		DIFF_FLAG_ISNT_SET(diff, GIT_DIFF_INCLUDE_UNREADABLE))
 		return 0;
@@ -706,6 +706,31 @@ static bool diff_time_eq(
 		(!use_nanos || a->nanoseconds == b->nanoseconds);
 }
 
+/*
+ * Test if the given index time is newer than the given existing index entry.
+ * If the timestamps are exactly equivalent, then the given index time is
+ * considered "racily newer" than the existing index entry.
+ */
+static bool diff_newer_than_index(
+	const git_index_time *a, const git_index *b, bool use_nanos)
+{
+	bool is_newer = false;
+
+	if(!b)
+		return false;
+
+	is_newer = is_newer || (a->seconds > (int32_t) b->stamp.mtime.tv_sec);
+	is_newer = is_newer || (!use_nanos &&
+		(a->seconds == (int32_t) b->stamp.mtime.tv_sec));
+	if(use_nanos)
+	{
+		is_newer = is_newer || ((a->seconds == (int32_t) b->stamp.mtime.tv_sec) &&
+			(a->nanoseconds >= (uint32_t) b->stamp.mtime.tv_nsec));
+	}
+
+	return is_newer;
+}
+
 typedef struct {
 	git_repository *repo;
 	git_iterator *old_iter;
@@ -838,7 +863,11 @@ static int maybe_modified(
 	 */
 	} else if (git_oid_iszero(&nitem->id) && new_is_workdir) {
 		bool use_ctime = ((diff->diffcaps & GIT_DIFFCAPS_TRUST_CTIME) != 0);
+#ifdef GIT_USE_NSEC
 		bool use_nanos = ((diff->diffcaps & GIT_DIFFCAPS_TRUST_NANOSECS) != 0);
+#else
+		bool use_nanos = false;
+#endif
 		git_index *index;
 		git_iterator_index(&index, info->new_iter);
 
@@ -863,7 +892,7 @@ static int maybe_modified(
 			oitem->ino != nitem->ino ||
 			oitem->uid != nitem->uid ||
 			oitem->gid != nitem->gid ||
-			(index && nitem->mtime.seconds >= index->stamp.mtime))
+			diff_newer_than_index(&nitem->mtime, index, use_nanos))
 		{
 			status = GIT_DELTA_MODIFIED;
 			modified_uncertain = true;
@@ -1260,7 +1289,18 @@ int git_diff__from_iterators(
 
 	/* run iterators building diffs */
 	while (!error && (info.oitem || info.nitem)) {
-		int cmp = info.oitem ?
+		int cmp;
+
+		/* report progress */
+		if (opts && opts->progress_cb) {
+			if ((error = opts->progress_cb(diff,
+					info.oitem ? info.oitem->path : NULL,
+					info.nitem ? info.nitem->path : NULL,
+					opts->payload)))
+				break;
+		}
+
+		cmp = info.oitem ?
 			(info.nitem ? diff->entrycomp(info.oitem, info.nitem) : -1) : 1;
 
 		/* create DELETED records for old items not matched in new */
